@@ -34,6 +34,7 @@ import tempfile
 import time
 import typing as T
 import xml.etree.ElementTree as ET
+import collections
 
 from mesonbuild import build
 from mesonbuild import environment
@@ -143,7 +144,7 @@ class InstalledFile:
                     p = p.with_name('{}-{}'.format(p.name, self.version[0]))
                 return p.with_suffix('.dll')
 
-            p = p.with_name('lib{}'.format(p.name))
+            p = p.with_name(f'lib{p.name}')
             if env.machines.host.is_darwin():
                 # MacOS only has libfoo.dylib and libfoo.X.dylib
                 if len(self.version) > 1:
@@ -179,7 +180,7 @@ class InstalledFile:
         elif self.typ == 'expr':
             return Path(platform_fix_name(p.as_posix(), canonical_compiler, env))
         else:
-            raise RuntimeError('Invalid installed file type {}'.format(self.typ))
+            raise RuntimeError(f'Invalid installed file type {self.typ}')
 
         return p
 
@@ -190,9 +191,9 @@ class InstalledFile:
         if self.typ == 'dir':
             abs_p = installdir / p
             if not abs_p.exists():
-                raise RuntimeError('{} does not exist'.format(p))
+                raise RuntimeError(f'{p} does not exist')
             if not abs_p.is_dir():
-                raise RuntimeError('{} is not a directory'.format(p))
+                raise RuntimeError(f'{p} is not a directory')
             return [x.relative_to(installdir) for x in abs_p.rglob('*') if x.is_file() or x.is_symlink()]
         else:
             return [p]
@@ -214,7 +215,7 @@ class TestDef:
 
     def display_name(self) -> str:
         if self.name:
-            return '{}   ({})'.format(self.path.as_posix(), self.name)
+            return f'{self.path.as_posix()}   ({self.name})'
         return self.path.as_posix()
 
     def __lt__(self, other: object) -> bool:
@@ -300,24 +301,24 @@ def validate_install(test: TestDef, installdir: Path, compiler: str, env: enviro
         try:
             expected_raw += i.get_paths(compiler, env, installdir)
         except RuntimeError as err:
-            ret_msg += 'Expected path error: {}\n'.format(err)
+            ret_msg += f'Expected path error: {err}\n'
     expected = {x: False for x in expected_raw}
     found = [x.relative_to(installdir) for x in installdir.rglob('*') if x.is_file() or x.is_symlink()]
     # Mark all found files as found and detect unexpected files
     for fname in found:
         if fname not in expected:
-            ret_msg += 'Extra file {} found.\n'.format(fname)
+            ret_msg += f'Extra file {fname} found.\n'
             continue
         expected[fname] = True
     # Check if expected files were found
     for p, f in expected.items():
         if not f:
-            ret_msg += 'Expected file {} missing.\n'.format(p)
+            ret_msg += f'Expected file {p} missing.\n'
     # List dir content on error
     if ret_msg != '':
         ret_msg += '\nInstall dir contents:\n'
         for i in found:
-            ret_msg += '  - {}\n'.format(i)
+            ret_msg += f'  - {i}\n'
     return ret_msg
 
 def log_text_file(logfile, testdir, stdo, stde):
@@ -370,15 +371,25 @@ def run_ci_commands(raw_log: str) -> T.List[str]:
         res += ['CI COMMAND {}:\n{}\n'.format(cmd[0], ci_commands[cmd[0]](cmd[1:]))]
     return res
 
+class OutputMatch:
+    def __init__(self, how: str, expected: str, count: int) -> None:
+        self.how = how
+        self.expected = expected
+        self.count = count
+
+    def match(self, actual: str) -> bool:
+        if self.how == "re":
+            return bool(re.match(self.expected, actual))
+        return self.expected == actual
+
 def _compare_output(expected: T.List[T.Dict[str, str]], output: str, desc: str) -> str:
     if expected:
-        i = iter(expected)
-
-        def next_expected(i):
-            # Get the next expected line
-            item = next(i)
+        matches = []
+        nomatches = []
+        for item in expected:
             how = item.get('match', 'literal')
             expected = item.get('line')
+            count = int(item.get('count', -1))
 
             # Simple heuristic to automatically convert path separators for
             # Windows:
@@ -396,23 +407,46 @@ def _compare_output(expected: T.List[T.Dict[str, str]], output: str, desc: str)
                     sub = r'\\\\'
                 expected = re.sub(r'/(?=.*(WARNING|ERROR))', sub, expected)
 
-            return how, expected
+            m = OutputMatch(how, expected, count)
+            if count == 0:
+                nomatches.append(m)
+            else:
+                matches.append(m)
 
-        try:
-            how, expected = next_expected(i)
-            for actual in output.splitlines():
-                if how == "re":
-                    match = bool(re.match(expected, actual))
+
+        i = 0
+        for actual in output.splitlines():
+            # Verify this line does not match any unexpected lines (item.count == 0)
+            for item in nomatches:
+                if item.match(actual):
+                    return f'unexpected "{item.expected}" found in {desc}'
+            # If we matched all expected lines, continue to verify there are
+            # no unexpected line. If nomatches is empty then we are done already.
+            if i >= len(matches):
+                if not nomatches:
+                    break
+                continue
+            # Check if this line match current expected line
+            item = matches[i]
+            if item.match(actual):
+                if item.count < 0:
+                    # count was not specified, continue with next expected line,
+                    # it does not matter if this line will be matched again or
+                    # not.
+                    i += 1
                 else:
-                    match = (expected == actual)
-                if match:
-                    how, expected = next_expected(i)
+                    # count was specified (must be >0), continue expecting this
+                    # same line. If count reached 0 we continue with next
+                    # expected line but remember that this one must not match
+                    # anymore.
+                    item.count -= 1
+                    if item.count == 0:
+                        nomatches.append(item)
+                        i += 1
 
+        if i < len(matches):
             # reached the end of output without finding expected
-            return 'expected "{}" not found in {}'.format(expected, desc)
-        except StopIteration:
-            # matched all expected lines
-            pass
+            return f'expected "{matches[i].expected}" not found in {desc}'
 
     return ''
 
@@ -461,6 +495,8 @@ def create_deterministic_builddir(test: TestDef, use_tmpdir: bool) -> str:
         src_dir += test.name
     rel_dirname = 'b ' + hashlib.sha256(src_dir.encode(errors='ignore')).hexdigest()[0:10]
     abs_pathname = os.path.join(tempfile.gettempdir() if use_tmpdir else os.getcwd(), rel_dirname)
+    if os.path.exists(abs_pathname):
+        mesonlib.windows_proof_rmtree(abs_pathname)
     os.mkdir(abs_pathname)
     return abs_pathname
 
@@ -474,7 +510,7 @@ def format_parameter_file(file_basename: str, test: TestDef, test_build_dir: str
 
     return destination
 
-def detect_parameter_files(test: TestDef, test_build_dir: str) -> (Path, Path):
+def detect_parameter_files(test: TestDef, test_build_dir: str) -> T.Tuple[Path, Path]:
     nativefile = test.path / 'nativefile.ini'
     crossfile = test.path / 'crossfile.ini'
 
@@ -486,7 +522,9 @@ def detect_parameter_files(test: TestDef, test_build_dir: str) -> (Path, Path):
 
     return nativefile, crossfile
 
-def run_test(test: TestDef, extra_args, compiler, backend, flags, commands, should_fail, use_tmp: bool):
+def run_test(test: TestDef, extra_args: T.List[str], compiler: str, backend: Backend,
+            flags: T.List[str], commands: T.Tuple[T.List[str], T.List[str], T.List[str], T.List[str]],
+            should_fail: bool, use_tmp: bool) -> T.Optional[TestResult]:
     if test.skip:
         return None
     build_dir = create_deterministic_builddir(test, use_tmp)
@@ -501,7 +539,10 @@ def run_test(test: TestDef, extra_args, compiler, backend, flags, commands, shou
     finally:
         mesonlib.windows_proof_rmtree(build_dir)
 
-def _run_test(test: TestDef, test_build_dir: str, install_dir: str, extra_args, compiler, backend, flags, commands, should_fail):
+def _run_test(test: TestDef, test_build_dir: str, install_dir: str,
+              extra_args: T.List[str], compiler: str, backend: Backend,
+              flags: T.List[str], commands: T.Tuple[T.List[str], T.List[str], T.List[str], T.List[str]],
+              should_fail: bool) -> TestResult:
     compile_commands, clean_commands, install_commands, uninstall_commands = commands
     gen_start = time.time()
     # Configure in-process
@@ -536,7 +577,7 @@ def _run_test(test: TestDef, test_build_dir: str, install_dir: str, extra_args,
         if returncode == 1:
             return testresult
         elif returncode != 0:
-            testresult.fail('Test exited with unexpected status {}.'.format(returncode))
+            testresult.fail(f'Test exited with unexpected status {returncode}.')
             return testresult
         else:
             testresult.fail('Test that should have failed succeeded.')
@@ -620,130 +661,142 @@ def _run_test(test: TestDef, test_build_dir: str, install_dir: str, extra_args,
 
     return testresult
 
-def gather_tests(testdir: Path, stdout_mandatory: bool) -> T.List[TestDef]:
-    tests = [t.name for t in testdir.iterdir() if t.is_dir()]
-    tests = [t for t in tests if not t.startswith('.')]  # Filter non-tests files (dot files, etc)
-    test_defs = [TestDef(testdir / t, None, []) for t in tests]
-    all_tests = []  # type: T.List[TestDef]
-    for t in test_defs:
-        test_def = {}
-        test_def_file = t.path / 'test.json'
-        if test_def_file.is_file():
-            test_def = json.loads(test_def_file.read_text())
 
-        # Handle additional environment variables
-        env = {}  # type: T.Dict[str, str]
-        if 'env' in test_def:
-            assert isinstance(test_def['env'], dict)
-            env = test_def['env']
-            for key, val in env.items():
-                val = val.replace('@ROOT@', t.path.resolve().as_posix())
-                val = val.replace('@PATH@', t.env.get('PATH', ''))
-                env[key] = val
+def load_test_json(t: TestDef, stdout_mandatory: bool) -> T.List[TestDef]:
+    all_tests: T.List[TestDef] = []
+    test_def = {}
+    test_def_file = t.path / 'test.json'
+    if test_def_file.is_file():
+        test_def = json.loads(test_def_file.read_text())
 
-        # Handle installed files
-        installed = []  # type: T.List[InstalledFile]
-        if 'installed' in test_def:
-            installed = [InstalledFile(x) for x in test_def['installed']]
+    # Handle additional environment variables
+    env = {}  # type: T.Dict[str, str]
+    if 'env' in test_def:
+        assert isinstance(test_def['env'], dict)
+        env = test_def['env']
+        for key, val in env.items():
+            val = val.replace('@ROOT@', t.path.resolve().as_posix())
+            val = val.replace('@PATH@', t.env.get('PATH', ''))
+            env[key] = val
 
-        # Handle expected output
-        stdout = test_def.get('stdout', [])
-        if stdout_mandatory and not stdout:
-            raise RuntimeError("{} must contain a non-empty stdout key".format(test_def_file))
+    # Handle installed files
+    installed = []  # type: T.List[InstalledFile]
+    if 'installed' in test_def:
+        installed = [InstalledFile(x) for x in test_def['installed']]
 
-        # Handle the do_not_set_opts list
-        do_not_set_opts = test_def.get('do_not_set_opts', [])  # type: T.List[str]
+    # Handle expected output
+    stdout = test_def.get('stdout', [])
+    if stdout_mandatory and not stdout:
+        raise RuntimeError(f"{test_def_file} must contain a non-empty stdout key")
 
-        # Skip tests if the tool requirements are not met
-        if 'tools' in test_def:
-            assert isinstance(test_def['tools'], dict)
-            for tool, vers_req in test_def['tools'].items():
-                if tool not in tool_vers_map:
-                    t.skip = True
-                elif not mesonlib.version_compare(tool_vers_map[tool], vers_req):
-                    t.skip = True
+    # Handle the do_not_set_opts list
+    do_not_set_opts = test_def.get('do_not_set_opts', [])  # type: T.List[str]
 
-        # Skip the matrix code and just update the existing test
-        if 'matrix' not in test_def:
-            t.env.update(env)
-            t.installed_files = installed
-            t.do_not_set_opts = do_not_set_opts
-            t.stdout = stdout
-            all_tests += [t]
-            continue
+    # Skip tests if the tool requirements are not met
+    if 'tools' in test_def:
+        assert isinstance(test_def['tools'], dict)
+        for tool, vers_req in test_def['tools'].items():
+            if tool not in tool_vers_map:
+                t.skip = True
+            elif not mesonlib.version_compare(tool_vers_map[tool], vers_req):
+                t.skip = True
 
-        # 'matrix; entry is present, so build multiple tests from matrix definition
-        opt_list = []  # type: T.List[T.List[T.Tuple[str, bool]]]
-        matrix = test_def['matrix']
-        assert "options" in matrix
-        for key, val in matrix["options"].items():
-            assert isinstance(val, list)
-            tmp_opts = []  # type: T.List[T.Tuple[str, bool]]
-            for i in val:
-                assert isinstance(i, dict)
-                assert "val" in i
-                skip = False
+    # Skip the matrix code and just update the existing test
+    if 'matrix' not in test_def:
+        t.env.update(env)
+        t.installed_files = installed
+        t.do_not_set_opts = do_not_set_opts
+        t.stdout = stdout
+        return [t]
 
-                # Skip the matrix entry if environment variable is present
-                if 'skip_on_env' in i:
-                    for skip_env_var in i['skip_on_env']:
-                        if skip_env_var in os.environ:
-                            skip = True
+    new_opt_list: T.List[T.List[T.Tuple[str, bool]]]
 
-                # Only run the test if all compiler ID's match
-                if 'compilers' in i:
-                    for lang, id_list in i['compilers'].items():
-                        if lang not in compiler_id_map or compiler_id_map[lang] not in id_list:
-                            skip = True
-                            break
+    # 'matrix; entry is present, so build multiple tests from matrix definition
+    opt_list = []  # type: T.List[T.List[T.Tuple[str, bool]]]
+    matrix = test_def['matrix']
+    assert "options" in matrix
+    for key, val in matrix["options"].items():
+        assert isinstance(val, list)
+        tmp_opts = []  # type: T.List[T.Tuple[str, bool]]
+        for i in val:
+            assert isinstance(i, dict)
+            assert "val" in i
+            skip = False
 
-                # Add an empty matrix entry
-                if i['val'] is None:
-                    tmp_opts += [(None, skip)]
-                    continue
+            # Skip the matrix entry if environment variable is present
+            if 'skip_on_env' in i:
+                for skip_env_var in i['skip_on_env']:
+                    if skip_env_var in os.environ:
+                        skip = True
 
-                tmp_opts += [('{}={}'.format(key, i['val']), skip)]
-
-            if opt_list:
-                new_opt_list = []  # type: T.List[T.List[T.Tuple[str, bool]]]
-                for i in opt_list:
-                    for j in tmp_opts:
-                        new_opt_list += [[*i, j]]
-                opt_list = new_opt_list
-            else:
-                opt_list = [[x] for x in tmp_opts]
-
-        # Exclude specific configurations
-        if 'exclude' in matrix:
-            assert isinstance(matrix['exclude'], list)
-            new_opt_list = []  # type: T.List[T.List[T.Tuple[str, bool]]]
-            for i in opt_list:
-                exclude = False
-                opt_names = [x[0] for x in i]
-                for j in matrix['exclude']:
-                    ex_list = ['{}={}'.format(k, v) for k, v in j.items()]
-                    if all([x in opt_names for x in ex_list]):
-                        exclude = True
+            # Only run the test if all compiler ID's match
+            if 'compilers' in i:
+                for lang, id_list in i['compilers'].items():
+                    if lang not in compiler_id_map or compiler_id_map[lang] not in id_list:
+                        skip = True
                         break
 
-                if not exclude:
-                    new_opt_list += [i]
+            # Add an empty matrix entry
+            if i['val'] is None:
+                tmp_opts += [(None, skip)]
+                continue
 
+            tmp_opts += [('{}={}'.format(key, i['val']), skip)]
+
+        if opt_list:
+            new_opt_list = []
+            for i in opt_list:
+                for j in tmp_opts:
+                    new_opt_list += [[*i, j]]
             opt_list = new_opt_list
+        else:
+            opt_list = [[x] for x in tmp_opts]
 
+    # Exclude specific configurations
+    if 'exclude' in matrix:
+        assert isinstance(matrix['exclude'], list)
+        new_opt_list = []
         for i in opt_list:
-            name = ' '.join([x[0] for x in i if x[0] is not None])
-            opts = ['-D' + x[0] for x in i if x[0] is not None]
-            skip = any([x[1] for x in i])
-            test = TestDef(t.path, name, opts, skip or t.skip)
-            test.env.update(env)
-            test.installed_files = installed
-            test.do_not_set_opts = do_not_set_opts
-            test.stdout = stdout
-            all_tests += [test]
+            exclude = False
+            opt_names = [x[0] for x in i]
+            for j in matrix['exclude']:
+                ex_list = [f'{k}={v}' for k, v in j.items()]
+                if all([x in opt_names for x in ex_list]):
+                    exclude = True
+                    break
 
+            if not exclude:
+                new_opt_list += [i]
+
+        opt_list = new_opt_list
+
+    for i in opt_list:
+        name = ' '.join([x[0] for x in i if x[0] is not None])
+        opts = ['-D' + x[0] for x in i if x[0] is not None]
+        skip = any([x[1] for x in i])
+        test = TestDef(t.path, name, opts, skip or t.skip)
+        test.env.update(env)
+        test.installed_files = installed
+        test.do_not_set_opts = do_not_set_opts
+        test.stdout = stdout
+        all_tests.append(test)
+
+    return all_tests
+
+
+def gather_tests(testdir: Path, stdout_mandatory: bool, only: T.List[str]) -> T.List[TestDef]:
+    all_tests: T.List[TestDef] = []
+    for t in testdir.iterdir():
+        # Filter non-tests files (dot files, etc)
+        if not t.is_dir() or t.name.startswith('.'):
+            continue
+        if only and not any(t.name.startswith(prefix) for prefix in only):
+            continue
+        test_def = TestDef(t, None, [])
+        all_tests.extend(load_test_json(test_def, stdout_mandatory))
     return sorted(all_tests)
 
+
 def have_d_compiler():
     if shutil.which("ldc2"):
         return True
@@ -800,7 +853,7 @@ def have_java():
         return True
     return False
 
-def skippable(suite, test):
+def skippable(suite: str, test: str) -> bool:
     # Everything is optional when not running on CI
     if not under_ci:
         return True
@@ -848,11 +901,6 @@ def skippable(suite, test):
     if test.endswith('34 gir static lib'):
         return True
 
-    # No frameworks test should be skipped on linux CI, as we expect all
-    # prerequisites to be installed
-    if mesonlib.is_linux():
-        return False
-
     # Boost test should only be skipped for windows CI build matrix entries
     # which don't define BOOST_ROOT
     if test.endswith('1 boost'):
@@ -860,14 +908,22 @@ def skippable(suite, test):
             return 'BOOST_ROOT' not in os.environ
         return False
 
-    # Qt is provided on macOS by Homebrew
-    if test.endswith('4 qt') and mesonlib.is_osx():
-        return False
+    # Not all OSes have all of the methods for qt (qmake and pkg-config), don't
+    # fail if that happens.
+    #
+    # On macOS we should have all of the requirements at all times.
+    if test.endswith('4 qt'):
+        return not mesonlib.is_osx()
 
     # Bindgen isn't available in all distros
     if test.endswith('12 bindgen'):
         return False
 
+    # No frameworks test should be skipped on linux CI, as we expect all
+    # prerequisites to be installed
+    if mesonlib.is_linux():
+        return False
+
     # Other framework tests are allowed to be skipped on other platforms
     return True
 
@@ -920,11 +976,11 @@ def should_skip_rust(backend: Backend) -> bool:
         return True
     return False
 
-def detect_tests_to_run(only: T.List[str], use_tmp: bool) -> T.List[T.Tuple[str, T.List[TestDef], bool]]:
+def detect_tests_to_run(only: T.Dict[str, T.List[str]], use_tmp: bool) -> T.List[T.Tuple[str, T.List[TestDef], bool]]:
     """
     Parameters
     ----------
-    only: list of str, optional
+    only: dict of categories and list of test cases, optional
         specify names of tests to run
 
     Returns
@@ -980,9 +1036,9 @@ def detect_tests_to_run(only: T.List[str], use_tmp: bool) -> T.List[T.Tuple[str,
     assert categories == ALL_TESTS, 'argparse("--only", choices=ALL_TESTS) need to be updated to match all_tests categories'
 
     if only:
-        all_tests = [t for t in all_tests if t.category in only]
+        all_tests = [t for t in all_tests if t.category in only.keys()]
 
-    gathered_tests = [(t.category, gather_tests(Path('test cases', t.subdir), t.stdout_mandatory), t.skip) for t in all_tests]
+    gathered_tests = [(t.category, gather_tests(Path('test cases', t.subdir), t.stdout_mandatory, only[t.category]), t.skip) for t in all_tests]
     return gathered_tests
 
 def run_tests(all_tests: T.List[T.Tuple[str, T.List[TestDef], bool]],
@@ -1039,7 +1095,7 @@ def _run_tests(all_tests: T.List[T.Tuple[str, T.List[TestDef], bool]],
             (testnum, testbase) = t.path.name.split(' ', 1)
             testname = '%.3d %s' % (int(testnum), testbase)
             if t.name:
-                testname += ' ({})'.format(t.name)
+                testname += f' ({t.name})'
             should_fail = False
             suite_args = []
             if name.startswith('failing'):
@@ -1067,7 +1123,7 @@ def _run_tests(all_tests: T.List[T.Tuple[str, T.List[TestDef], bool]],
             else:
                 without_install = "" if len(install_commands) > 0 else " (without install)"
                 if result.msg != '':
-                    print(red('Failed test{} during {}: {!r}'.format(without_install, result.step.name, t.display_name())))
+                    print(red(f'Failed test{without_install} during {result.step.name}: {t.display_name()!r}'))
                     print('Reason:', result.msg)
                     failing_tests += 1
                     if result.step == BuildStep.configure and result.mlog != no_meson_log_msg:
@@ -1091,7 +1147,7 @@ def _run_tests(all_tests: T.List[T.Tuple[str, T.List[TestDef], bool]],
                         for (_, _, res) in futures:
                             res.cancel()
                 else:
-                    print('Succeeded test%s: %s' % (without_install, t.display_name()))
+                    print(f'Succeeded test{without_install}: {t.display_name()}')
                     passing_tests += 1
                 conf_time += result.conftime
                 build_time += result.buildtime
@@ -1173,26 +1229,26 @@ def check_meson_commands_work(options):
         gen_cmd = meson_commands + [testdir, build_dir] + backend_flags + options.extra_args
         pc, o, e = Popen_safe(gen_cmd)
         if pc.returncode != 0:
-            raise RuntimeError('Failed to configure {!r}:\n{}\n{}'.format(testdir, e, o))
+            raise RuntimeError(f'Failed to configure {testdir!r}:\n{e}\n{o}')
         print('Checking that introspect works...')
         pc, o, e = Popen_safe(meson_commands + ['introspect', '--targets'], cwd=build_dir)
         json.loads(o)
         if pc.returncode != 0:
-            raise RuntimeError('Failed to introspect --targets {!r}:\n{}\n{}'.format(testdir, e, o))
+            raise RuntimeError(f'Failed to introspect --targets {testdir!r}:\n{e}\n{o}')
         print('Checking that building works...')
         dir_args = get_backend_args_for_dir(backend, build_dir)
         pc, o, e = Popen_safe(compile_commands + dir_args, cwd=build_dir)
         if pc.returncode != 0:
-            raise RuntimeError('Failed to build {!r}:\n{}\n{}'.format(testdir, e, o))
+            raise RuntimeError(f'Failed to build {testdir!r}:\n{e}\n{o}')
         print('Checking that testing works...')
         pc, o, e = Popen_safe(test_commands, cwd=build_dir)
         if pc.returncode != 0:
-            raise RuntimeError('Failed to test {!r}:\n{}\n{}'.format(testdir, e, o))
+            raise RuntimeError(f'Failed to test {testdir!r}:\n{e}\n{o}')
         if install_commands:
             print('Checking that installing works...')
             pc, o, e = Popen_safe(install_commands, cwd=build_dir)
             if pc.returncode != 0:
-                raise RuntimeError('Failed to install {!r}:\n{}\n{}'.format(testdir, e, o))
+                raise RuntimeError(f'Failed to install {testdir!r}:\n{e}\n{o}')
 
 
 def detect_system_compiler(options):
@@ -1229,7 +1285,7 @@ def detect_system_compiler(options):
 
 def print_compilers(env, machine):
     print()
-    print('{} machine compilers'.format(machine.get_lower_case_name()))
+    print(f'{machine.get_lower_case_name()} machine compilers')
     print()
     for lang in sorted(compilers.all_languages):
         try:
@@ -1237,7 +1293,7 @@ def print_compilers(env, machine):
             details = '{:<10} {} {}'.format('[' + comp.get_id() + ']', ' '.join(comp.get_exelist()), comp.get_version_string())
         except mesonlib.MesonException:
             details = '[not found]'
-        print('%-7s: %s' % (lang, details))
+        print(f'{lang:<7}: {details}')
 
 
 def print_tool_versions():
@@ -1278,7 +1334,7 @@ def print_tool_versions():
                 tool_vers_map[t['tool']] = m.group(t['match_group'])
                 return '{} ({})'.format(exe, m.group(t['match_group']))
 
-        return '{} (unknown)'.format(exe)
+        return f'{exe} (unknown)'
 
     print()
     print('tools')
@@ -1289,6 +1345,14 @@ def print_tool_versions():
         print('{0:<{2}}: {1}'.format(tool['tool'], get_version(tool), max_width))
     print()
 
+def clear_transitive_files():
+    a = Path('test cases/common')
+    for d in a.glob('*subproject subdir/subprojects/subsubsub*'):
+        if d.is_dir():
+            mesonlib.windows_proof_rmtree(d)
+        else:
+            mesonlib.windows_proof_rm(str(d))
+
 if __name__ == '__main__':
     parser = argparse.ArgumentParser(description="Run the test suite of Meson.")
     parser.add_argument('extra_args', nargs='*',
@@ -1298,7 +1362,8 @@ if __name__ == '__main__':
                         help='Stop running if test case fails')
     parser.add_argument('--no-unittests', action='store_true',
                         help='Not used, only here to simplify run_tests.py')
-    parser.add_argument('--only', help='name of test(s) to run', nargs='+', choices=ALL_TESTS)
+    parser.add_argument('--only', default=[],
+                        help='name of test(s) to run, in format "category[/name]" where category is one of: ' + ', '.join(ALL_TESTS), nargs='+')
     parser.add_argument('--cross-file', action='store', help='File describing cross compilation environment.')
     parser.add_argument('--native-file', action='store', help='File describing native compilation environment.')
     parser.add_argument('--use-tmpdir', action='store_true', help='Use tmp directory for temporary files.')
@@ -1309,6 +1374,8 @@ if __name__ == '__main__':
     if options.native_file:
         options.extra_args += ['--native-file', options.native_file]
 
+    clear_transitive_files()
+
     print('Meson build system', meson_version, 'Project Tests')
     print('Using python', sys.version.split('\n')[0])
     if 'VSCMD_VER' in os.environ:
@@ -1321,8 +1388,15 @@ if __name__ == '__main__':
         os.chdir(script_dir)
     check_format()
     check_meson_commands_work(options)
+    only = collections.defaultdict(list)
+    for i in options.only:
+        try:
+            cat, case = i.split('/')
+            only[cat].append(case)
+        except ValueError:
+            only[i].append('')
     try:
-        all_tests = detect_tests_to_run(options.only, options.use_tmpdir)
+        all_tests = detect_tests_to_run(only, options.use_tmpdir)
         (passing_tests, failing_tests, skipped_tests) = run_tests(all_tests, 'meson-test-run', options.failfast, options.extra_args, options.use_tmpdir)
     except StopException:
         pass
@@ -1337,9 +1411,10 @@ if __name__ == '__main__':
             except UnicodeError:
                 print(l.encode('ascii', errors='replace').decode(), '\n')
     for name, dirs, _ in all_tests:
-        dir_names = list(set(x.path.name for x in dirs))
+        dir_names = list({x.path.name for x in dirs})
         for k, g in itertools.groupby(dir_names, key=lambda x: x.split()[0]):
             tests = list(g)
             if len(tests) != 1:
-                print('WARNING: The %s suite contains duplicate "%s" tests: "%s"' % (name, k, '", "'.join(tests)))
+                print('WARNING: The {} suite contains duplicate "{}" tests: "{}"'.format(name, k, '", "'.join(tests)))
+    clear_transitive_files()
     raise SystemExit(failing_tests)
